#!/usr/bin/env python3
"""Hub functions."""

import sys
from pathlib import Path

from tolkein import tofile
from tolkein import tolog

from .attributes import fetch_types
from .hub import process_names_file

LOGGER = tolog.logger(__name__)


def validate_types_file(types_file, dir_path, es, types_name, opts, *, attributes=None):
    """Validate types file."""
    LOGGER.info("Validating YAML file %s", types_file)
    try:
        if not isinstance(types_file, str):
            types_file = str(types_file.resolve())
        types = tofile.load_yaml(types_file)
    except Exception:
        LOGGER.error("Unable to open types file %s", types_file)
        sys.exit(1)
    if "attributes" in types:
        if attributes is None:
            attributes = {}
        try:
            stored_attributes = fetch_types(es, types_name, opts)
            attributes = {**stored_attributes, **attributes}
        except Exception:
            pass
        if attributes:
            sequence = max(int(d["sequence"]) for d in attributes.values()) + 1
        else:
            sequence = 0
        for key, entry in types["attributes"].items():
            if not isinstance(entry, dict):
                entry = {"default": entry}
            if key in attributes:
                entry = {**attributes[key], **entry}
            if "sequence" not in entry:
                entry["sequence"] = sequence
                sequence += 1
            types["attributes"][key] = entry
            enum = entry.get("constraint", {}).get("enum", [])
            if enum:
                entry["constraint"]["enum"] = [value.lower() for value in enum]
            translate = entry.get("translate", {})
            if translate:
                entry["translate"] = {
                    key.lower(): value for key, value in translate.items()
                }
    names = None
    exclusions = None
    data = None
    if "file" in types:
        if "name" in types["file"]:
            if "taxonomy" not in types and (
                "features" not in types or "taxon_id" not in types["features"]
            ):
                LOGGER.error("Types file contains no taxonomy information")
                sys.exit(1)
            datafile = Path(dir_path) / types["file"]["name"]
            data = tofile.open_file_handle(datafile)
            if data is None:
                LOGGER.error("Data file '%s' could not de opened for reading", datafile)
                sys.exit(1)
        defaults = {"attributes": {}, "metadata": {}}
        if "defaults" in types:
            for key, value in types["defaults"].items():
                defaults[key] = value
        for key, value in types["file"].items():
            if key.startswith("display") or key.startswith("taxon"):
                defaults["attributes"].update({key: value})
            elif key.startswith("source"):
                defaults["attributes"].update({key: value})
                defaults["metadata"].update({key: value})
        types.update({"defaults": defaults})
        if "name" in types["file"]:
            names = process_names_file(
                types, Path(dir_path) / "names" / types["file"]["name"]
            )
            if "exclusions" in types["file"]:
                exclusions = process_names_file(
                    types,
                    Path(dir_path) / "exclusions" / types["file"]["name"],
                    value_path=types["file"]["exclusions"],
                )
    return types, data, names, exclusions



"""
A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

import io
import re
from os.path import dirname
from os.path import join

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    """Read file."""
    with io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ) as fh:
        return fh.read()


setup(
    name="genomehubs",  # Required
    version="2.10.15",
    description="GenomeHubs",  # Optional
    long_description="%s\n%s"
    % (
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub(
            "", read("README.rst")
        ),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
    ),
    long_description_content_type="text/x-rst",  # Optional (see note above)
    url="https://github.com/genomehubs/genomehubs",  # Optional
    # This should be your name or the name of the organization which owns the
    # project.
    author="genomehubs",  # Optional
    # This should be a valid email address corresponding to the author listed
    # above.
    author_email="genomehubs@genomehubs.org",  # Optional
    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ],
    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a list of additional keywords, separated
    # by commas, to be used to assist searching for the distribution in a
    # larger catalog.
    keywords="bioinformatics",  # Optional
    # When your source code is in a subdirectory under the project root, e.g.
    # `src/`, it is necessary to specify the `package_dir` argument.
    package_dir={"": "src"},  # Optional
    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(where="src"),  # Required
    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. See
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires=">=3.6, <4",
    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "biopython>=1.78",
        "boto3>=1.34.84",
        "docopt>=0.6.2",
        "elasticsearch==8.7",
        "fastjsonschema>=2.15.3",
        "filetype>=1.0.7",
        "Pillow>=8.0",
        "pyyaml",
        "sparqlwrapper>=1.4.1",
        "tolkein>=0.5.0",
        "ujson>=3.0.0",
        "urllib3>=1.26.2, <2",  # for elastic-transport
    ],  # Optional
    setup_requires=[
        "wheel",
    ],
    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={  # Optional
        "dev": ["pycodestyle>=2.6.0", "pydocstyle>=5.0.2", "pylint>=2.5.3"],
        "test": [
            "coverage>=5.1",
            "coveralls>=2.0.0",
            "mock>=4.0.2",
            "pytest-cov>=2.10.0",
            "pytest-isort>=1.1.0",
            "pytest-mock>=3.1.1",
            "pytest>=6.0.0",
        ],
    },
    entry_points={
        "console_scripts": ["genomehubs = genomehubs:cli"],
        "genomehubs.subcmd": [
            "fill = genomehubs.lib.fill:cli",
            "index = genomehubs.lib.index:cli",
            "init = genomehubs.lib.init:cli",
            "parse = genomehubs.lib.parse:cli",
            "run = genomehubs.lib.run:cli",
            "test = genomehubs.lib.test:cli",
        ],
    },
    project_urls={
        "Bug Reports": "https://github.com/genomehubs/genomehubs/issues",
        "Source": "https://github.com/genomehubs/genomehubs",
    },
    package_data={
        "genomehubs": ["genomehubs/config/dist.config.yaml", "genomehubs/py.typed"]
    },
    include_package_data=True,
)



#!/usr/bin/env python3
"""BUSCO functions."""

import glob
import os
import shlex
import subprocess

from tolkein import tofile
from tolkein import tolog

from .busco import parse_busco_lineages


LOGGER = tolog.logger(__name__)


def prepare_window_files(btk_dir, accession, prefix, opts):
    """Generate tsv/yaml pairs for windowed assembly statistics."""
    windows = "--window " + " --window ".join(opts["window-size"])
    out_dir = os.path.dirname(os.path.abspath(opts["outfile"]))
    out_dir = out_dir.replace("assembly", "feature")
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    outfile = f"{out_dir}/{accession}.window_stats.tsv"
    if not os.path.exists(outfile):
        cmd = f"btk pipeline window-stats {windows} \
        --min-window-length 1000 \
        --min-window-count 5 \
        --in {btk_dir}/{prefix}.chunk_stats.tsv.gz \
        --out {outfile}"
        print(cmd)
        subprocess.run(shlex.split(cmd))
    for window in opts["window-size"]:
        flag = "--window"
        if window == "1":
            flag += "-full"
            window_file = outfile
        else:
            window_file = f"{out_dir}/{accession}.window_stats.{window}.tsv"
        if os.path.exists(window_file):
            cmd = f"genomehubs parse {flag} {btk_dir} \
            --outfile {outfile}"
            if window != "1":
                cmd += f" --window-size {window}"
            print(cmd)
            subprocess.run(shlex.split(cmd))


def prepare_busco_features(btk_dir, lineages, accession, prefix, opts):
    """Parse busco full table as features."""
    """Generate tsv/yaml pairs for windowed assembly statistics."""
    out_dir = os.path.dirname(os.path.abspath(opts["outfile"]))
    out_dir = out_dir.replace("assembly", "feature")
    for lineage in lineages:
        outfile = f"{out_dir}/{accession}.busco.{lineage}.tsv"
        cmd = f"genomehubs parse \
        --busco-feature {btk_dir}/{prefix}.busco.{lineage}_odb10/{prefix}.busco.{lineage}_odb10/full_table.tsv.gz \
        --outfile {outfile} --config {btk_dir}/config.yaml"
        print(cmd)
        subprocess.run(shlex.split(cmd))


def directory_parser(params, opts, *, types=None, names=None):
    """Parse BlobToolKit data directory."""
    parsed = []
    template_keys = set()
    expanded_keys = set()
    for d in glob.glob(f"{opts['directory']}/*/", recursive=False):
        d = os.path.abspath(d)
        if os.path.exists(f"{d}/config.yaml"):
            config = tofile.load_yaml(f"{d}/config.yaml")
            accession = config["assembly"]["accession"]
            prefix = config["assembly"]["prefix"]
            if "window-size" in opts:
                prepare_window_files(d, accession, prefix, opts)
            taxid = config["taxon"]["taxid"]
            prefix = config["assembly"]["prefix"]
            parsed_row = {"taxon_id": taxid, "assembly_id": accession}
            lineages = parse_busco_lineages(
                types, template_keys, expanded_keys, d, config, prefix, parsed_row
            )
            prepare_busco_features(d, lineages, accession, prefix, opts)
        parsed.append(parsed_row)
    for key in template_keys:
        del types["attributes"][key]
    return parsed



#!/usr/bin/env python3

"""
Test index and API responses.

Usage:
    genomehubs test [--taxonomy-source STRING] [--config-file YAML...]
                    [--base-url URL] --json-test-dir DIR...

Options:
    --config-file YAML          YAML configuration file.
    --base-url URL             GenomeHubs API base url. [Default: http://localhost:3000/api/v2]
    --json-test-dir DIR        Directory containing templates for json response tests.
    -h, --help                 Show this
    -v, --version              Show version number
"""


import json
import os
import re
import sys
from textwrap import indent
from urllib.error import HTTPError
from urllib.parse import quote
from urllib.parse import unquote
from urllib.request import Request
from urllib.request import urlopen

import yaml
from docopt import docopt
from fastjsonschema import validate

from .config import config
from .version import __version__


def is_subset(a, b, path=None):
    """Test if a is a subset of b."""
    if path is None:
        path = {}

    def deep_unique(arr):
        """Test all items in b for uniqueness."""
        if isinstance(arr, list) and len(arr) > 1:
            for index, x in enumerate(arr[:-1]):
                for y in arr[index + 1:]:
                    if x == y:
                        return False
        return True

    def recursive_check_subset(a, b, path, *, invert=False):
        if isinstance(a, str):
            if a in b:
                path[a] = "FAIL" if invert else "PASS"
                yield True
            else:
                path[a] = "PASS" if invert else "FAIL"
                yield False
            return
        for x in a:
            if x == "not":
                path[x] = {}
                subset = next(recursive_check_subset(a[x], b, path[x], invert=True))
                yield not subset
            if x == "uniqueItems":
                if a[x] is True:
                    path[x] = {}
                    unique = deep_unique(b)
                    if invert:
                        unique = not unique
                    path[x] = "PASS" if unique else "FAIL"
                yield unique
            elif x == "jsonSchema":
                try:
                    validate(a[x], b)
                    path[x] = "FAIL" if invert else "PASS"
                    yield True
                except Exception:
                    path[x] = "PASS" if invert else "FAIL"
                    yield False
            elif isinstance(a[x], dict):
                if x in b:
                    path[x] = {}
                    yield all(recursive_check_subset(a[x], b[x], path[x]))
                else:
                    path[x] = False
                    yield path[x]
            else:
                if x in b and isinstance(a[x], list):
                    path[x] = []
                    matches = []
                    for i, y in enumerate(a[x]):
                        possible_path = {}
                        path[x].append(possible_path)
                        has_match = False
                        for z in b[x]:
                            path[x][i] = possible_path
                            if is_subset(y, z, path[x][i]):
                                has_match = True
                                break
                        matches.append(has_match)
                    yield all(matches)
                elif x == "anyOf":
                    path[x] = {}
                    yield any([is_subset(entry, b, path[x]) for entry in a[x]])
                else:
                    status = (x in b) and (a[x] == b[x])
                    if invert:
                        path[x] = "FAIL" if status else "PASS"
                    else:
                        path[x] = "PASS" if status else "FAIL"
                    yield status

    return all(recursive_check_subset(a, b, path))


def placeholder_substitution(string, config):
    """Substitute placeholders with config values."""
    parts = re.split(r"(?:\{\{|\}\})", string)
    replaced = []
    for index, part in enumerate(parts):
        if index % 2 == 1:
            try:
                replaced.append(config[part])
            except KeyError:
                print("ERROR: no value for {{%s}} in configuration" % part)
        else:
            replaced.append(part)
    return "".join(replaced)


def json_response_tests(base_url, template_dir, opts):
    """Check yaml template is a subset of remote JSON object."""
    required_headers = {"assert", "endpoint"}
    files = []
    for file in os.listdir(template_dir):
        if file.endswith((".yaml", ".yml")):
            files.append(file)
    success = True
    for file in files:
        with open("%s/%s" % (template_dir, file), "r") as stream:
            try:
                test = yaml.safe_load(stream)
            except yaml.YAMLError as err:
                print(err)
                exit(1)
            raw_query = test.get("querystring", None)
            if required_headers.issubset(set(test.keys())):
                if raw_query is not None:
                    qs = {
                        unquote(entry.split("=")[0]): unquote(entry.split("=")[1])
                        for entry in test.get("querystring", "_").split("&")
                    }
                    querystring = "&".join(
                        [
                            "%s=%s" % (quote(key), quote(value))
                            for key, value in qs.items()
                        ]
                    )
                else:
                    querystring = ""
                endpoint = placeholder_substitution(test["endpoint"], opts)
                url = "%s/%s?%s" % (base_url, endpoint, querystring)
                req = Request(url)
                req.add_header("accept", "application/json")
                try:
                    content = urlopen(req).read()
                    data = json.loads(content)
                except HTTPError:
                    success = False
                    print("    - %s: FAIL" % file)
                    print("      Unable to open %s" % url)
                    continue
                template = test["assert"]
                path = {}
                try:
                    assert is_subset(template, data, path)
                    print("    - %s: PASS" % file)
                except AssertionError:
                    print("    - %s: FAIL" % file)
                    success = False
                    print("      Test output:")
                    print(indent(yaml.dump(path), "          "))
                    print(data)
    return success


def test_json_dir(json_dirs, base_url, opts):
    """Run tests in a JSON test dir."""
    if not base_url.startswith("http"):
        base_url = "http://%s" % base_url
    summary = {}
    global_outcome = "PASS"
    print("Running tests:")
    if json_dirs is not None:
        if not isinstance(json_dirs, list):
            json_dirs = [json_dirs]
        for json_dir in json_dirs:
            print(indent(json_dir, "  "))
            success = json_response_tests(base_url, json_dir, opts)
            summary[json_dir] = success
    print()
    print("Summary:")
    for dir, success in summary.items():
        if success is True:
            outcome = "PASS"
        else:
            outcome = "FAIL"
            global_outcome = "FAIL"
        print("  %s: %s" % (dir, outcome))
    print()
    print("Outcome: %s" % global_outcome)
    if global_outcome == "FAIL":
        return False
    return True


def main(args):
    """Entry point."""
    options = config("test", **args)
    json_dirs = args.get("--json-test-dir", None)
    result = test_json_dir(json_dirs, args["--base-url"], options["test"])

    if result is False:
        exit(1)


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index("test") + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    main(args)


if __name__ == "__main__":
    cli()
